package org.eclipse.jdt.internal.compiler.flow;

package .tmp;

import java.util.ArrayList;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.AbstractMethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.Expression;
import org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable;
import org.eclipse.jdt.internal.compiler.ast.LabeledStatement;
import org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching;
import org.eclipse.jdt.internal.compiler.ast.Reference;
import org.eclipse.jdt.internal.compiler.ast.SubRoutineStatement;
import org.eclipse.jdt.internal.compiler.ast.ThrowStatement;
import org.eclipse.jdt.internal.compiler.ast.TryStatement;
import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;
import org.eclipse.jdt.internal.compiler.codegen.BranchLabel;
import org.eclipse.jdt.internal.compiler.flow.ExceptionHandlingFlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.flow.InitializationFlowContext;
import org.eclipse.jdt.internal.compiler.flow.LoopingFlowContext;
import org.eclipse.jdt.internal.compiler.flow.UnconditionalFlowInfo;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.CatchParameterBinding;
import org.eclipse.jdt.internal.compiler.lookup.LocalVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;
import org.eclipse.jdt.internal.compiler.lookup.VariableBinding;

public class FlowContext implements TypeConstants {
  public static final org.eclipse.jdt.internal.compiler.flow.FlowContext NotContinuableContext = new org.eclipse.jdt.internal.compiler.flow.FlowContext(null, null);
  
  public ASTNode associatedNode;
  
  public org.eclipse.jdt.internal.compiler.flow.FlowContext parent;
  
  public FlowInfo initsOnFinally;
  
  public int conditionalLevel = -1;
  
  public int tagBits;
  
  public TypeBinding[][] providedExpectedTypes = null;
  
  private Reference[] nullCheckedFieldReferences = null;
  
  private int[] timesToLiveForNullCheckInfo = null;
  
  public static final int DEFER_NULL_DIAGNOSTIC = 1;
  
  public static final int PREEMPT_NULL_DIAGNOSTIC = 2;
  
  public static final int INSIDE_NEGATION = 4;
  
  public static final int HIDE_NULL_COMPARISON_WARNING = 4096;
  
  public static final int HIDE_NULL_COMPARISON_WARNING_MASK = 61440;
  
  public static final int CAN_ONLY_NULL_NON_NULL = 0;
  
  public static final int CAN_ONLY_NULL = 1;
  
  public static final int CAN_ONLY_NON_NULL = 2;
  
  public static final int MAY_NULL = 3;
  
  public static final int ASSIGN_TO_NONNULL = 128;
  
  public static final int IN_UNBOXING = 16;
  
  public static final int EXIT_RESOURCE = 2048;
  
  public static final int CHECK_MASK = 255;
  
  public static final int IN_COMPARISON_NULL = 256;
  
  public static final int IN_COMPARISON_NON_NULL = 512;
  
  public static final int IN_ASSIGNMENT = 768;
  
  public static final int IN_INSTANCEOF = 1024;
  
  public static final int CONTEXT_MASK = -61696;
  
  public FlowContext(org.eclipse.jdt.internal.compiler.flow.FlowContext parent, ASTNode associatedNode) {
    this.parent = parent;
    this.associatedNode = associatedNode;
    if (parent != null) {
      if ((parent.tagBits & 0x3) != 0)
        this.tagBits |= 0x1; 
      this.initsOnFinally = parent.initsOnFinally;
      this.conditionalLevel = parent.conditionalLevel;
      this.nullCheckedFieldReferences = parent.nullCheckedFieldReferences;
      this.timesToLiveForNullCheckInfo = parent.timesToLiveForNullCheckInfo;
    } 
  }
  
  public void recordNullCheckedFieldReference(Reference reference, int timeToLive) {
    if (this.nullCheckedFieldReferences == null) {
      this.nullCheckedFieldReferences = new Reference[] { reference };
      this.timesToLiveForNullCheckInfo = new int[] { timeToLive, -1 };
    } else {
      int len = this.nullCheckedFieldReferences.length;
      for (int i = 0; i < len; i++) {
        if (this.nullCheckedFieldReferences[i] == null) {
          this.nullCheckedFieldReferences[i] = reference;
          this.timesToLiveForNullCheckInfo[i] = timeToLive;
          return;
        } 
      } 
      System.arraycopy(this.nullCheckedFieldReferences, 0, this.nullCheckedFieldReferences = new Reference[len + 2], 0, len);
      System.arraycopy(this.timesToLiveForNullCheckInfo, 0, this.timesToLiveForNullCheckInfo = new int[len + 2], 0, len);
      this.nullCheckedFieldReferences[len] = reference;
      this.timesToLiveForNullCheckInfo[len] = timeToLive;
    } 
  }
  
  public void extendTimeToLiveForNullCheckedField(int t) {
    if (this.timesToLiveForNullCheckInfo != null)
      for (int i = 0; i < this.timesToLiveForNullCheckInfo.length; i++) {
        if (this.timesToLiveForNullCheckInfo[i] > 0)
          this.timesToLiveForNullCheckInfo[i] = this.timesToLiveForNullCheckInfo[i] + t; 
      }  
  }
  
  public void expireNullCheckedFieldInfo() {
    if (this.nullCheckedFieldReferences != null)
      for (int i = 0; i < this.nullCheckedFieldReferences.length; i++) {
        this.timesToLiveForNullCheckInfo[i] = this.timesToLiveForNullCheckInfo[i] - 1;
        if (this.timesToLiveForNullCheckInfo[i] - 1 == 0)
          this.nullCheckedFieldReferences[i] = null; 
      }  
  }
  
  public boolean isNullcheckedFieldAccess(Reference reference) {
    if (this.nullCheckedFieldReferences == null)
      return false; 
    int len = this.nullCheckedFieldReferences.length;
    for (int i = 0; i < len; i++) {
      Reference checked = this.nullCheckedFieldReferences[i];
      if (checked != null)
        if (checked.isEquivalent(reference))
          return true;  
    } 
    return false;
  }
  
  public BranchLabel breakLabel() {
    return null;
  }
  
  public void checkExceptionHandlers(TypeBinding raisedException, ASTNode location, FlowInfo flowInfo, BlockScope scope) {
    checkExceptionHandlers(raisedException, location, flowInfo, scope, false);
  }
  
  public void checkExceptionHandlers(TypeBinding raisedException, ASTNode location, FlowInfo flowInfo, BlockScope scope, boolean isExceptionOnAutoClose) {
    org.eclipse.jdt.internal.compiler.flow.FlowContext traversedContext = this;
    ArrayList<LoopingFlowContext> abruptlyExitedLoops = null;
    if ((scope.compilerOptions()).sourceLevel >= 3342336L && location instanceof ThrowStatement) {
      Expression throwExpression = ((ThrowStatement)location).exception;
      LocalVariableBinding throwArgBinding = throwExpression.localVariableBinding();
      if (throwExpression instanceof org.eclipse.jdt.internal.compiler.ast.SingleNameReference && 
        throwArgBinding instanceof CatchParameterBinding && throwArgBinding.isEffectivelyFinal()) {
        CatchParameterBinding parameter = (CatchParameterBinding)throwArgBinding;
        checkExceptionHandlers(parameter.getPreciseTypes(), location, flowInfo, scope);
        return;
      } 
    } 
    while (traversedContext != null) {
      SubRoutineStatement sub;
      if ((sub = traversedContext.subroutine()) != null && sub.isSubRoutineEscaping())
        return; 
      if (traversedContext instanceof ExceptionHandlingFlowContext) {
        ExceptionHandlingFlowContext exceptionContext = 
          (ExceptionHandlingFlowContext)traversedContext;
        ReferenceBinding[] caughtExceptions;
        if ((caughtExceptions = exceptionContext.handledExceptions) != Binding.NO_EXCEPTIONS) {
          boolean definitelyCaught = false;
          for (int caughtIndex = 0, caughtCount = caughtExceptions.length; 
            caughtIndex < caughtCount; 
            caughtIndex++) {
            UnconditionalFlowInfo unconditionalFlowInfo;
            ReferenceBinding caughtException = caughtExceptions[caughtIndex];
            FlowInfo exceptionFlow = flowInfo;
            int state = (caughtException == null) ? 
              -1 : 
              Scope.compareTypes(raisedException, (TypeBinding)caughtException);
            if (abruptlyExitedLoops != null && caughtException != null && state != 0) {
              for (int i = 0, abruptlyExitedLoopsCount = abruptlyExitedLoops.size(); i < abruptlyExitedLoopsCount; i++) {
                LoopingFlowContext loop = abruptlyExitedLoops.get(i);
                loop.recordCatchContextOfEscapingException(exceptionContext, caughtException, flowInfo);
              } 
              unconditionalFlowInfo = FlowInfo.DEAD_END;
            } 
            switch (state) {
              case -1:
                exceptionContext.recordHandlingException(
                    caughtException, 
                    unconditionalFlowInfo.unconditionalInits(), 
                    raisedException, 
                    raisedException, 
                    location, 
                    definitelyCaught);
                definitelyCaught = true;
                break;
              case 1:
                exceptionContext.recordHandlingException(
                    caughtException, 
                    unconditionalFlowInfo.unconditionalInits(), 
                    raisedException, 
                    (TypeBinding)caughtException, 
                    location, 
                    false);
                break;
            } 
          } 
          if (definitelyCaught)
            return; 
        } 
        if (exceptionContext.isMethodContext) {
          if (raisedException.isUncheckedException(false))
            return; 
          boolean shouldMergeUnhandledExceptions = exceptionContext instanceof org.eclipse.jdt.internal.compiler.flow.ExceptionInferenceFlowContext;
          if (exceptionContext.associatedNode instanceof AbstractMethodDeclaration) {
            AbstractMethodDeclaration method = (AbstractMethodDeclaration)exceptionContext.associatedNode;
            if (method.isConstructor() && method.binding.declaringClass.isAnonymousType())
              shouldMergeUnhandledExceptions = true; 
          } 
          if (shouldMergeUnhandledExceptions) {
            exceptionContext.mergeUnhandledException(raisedException);
            return;
          } 
          break;
        } 
      } else if (traversedContext instanceof LoopingFlowContext) {
        if (abruptlyExitedLoops == null)
          abruptlyExitedLoops = new ArrayList<>(5); 
        abruptlyExitedLoops.add(traversedContext);
      } 
      traversedContext.recordReturnFrom(flowInfo.unconditionalInits());
      if (!isExceptionOnAutoClose && 
        traversedContext instanceof org.eclipse.jdt.internal.compiler.flow.InsideSubRoutineFlowContext) {
        ASTNode node = traversedContext.associatedNode;
        if (node instanceof TryStatement) {
          TryStatement tryStatement = (TryStatement)node;
          flowInfo.addInitializationsFrom((FlowInfo)tryStatement.subRoutineInits);
        } 
      } 
      traversedContext = traversedContext.getLocalParent();
    } 
    if (isExceptionOnAutoClose) {
      scope.problemReporter().unhandledExceptionFromAutoClose(raisedException, location);
    } else {
      scope.problemReporter().unhandledException(raisedException, location);
    } 
  }
  
  public void checkExceptionHandlers(TypeBinding[] raisedExceptions, ASTNode location, FlowInfo flowInfo, BlockScope scope) {
    int raisedCount;
    if (raisedExceptions == null || (
      raisedCount = raisedExceptions.length) == 0)
      return; 
    int remainingCount = raisedCount;
    System.arraycopy(
        raisedExceptions, 
        0, 
        raisedExceptions = new TypeBinding[raisedCount], 
        0, 
        raisedCount);
    org.eclipse.jdt.internal.compiler.flow.FlowContext traversedContext = this;
    ArrayList<LoopingFlowContext> abruptlyExitedLoops = null;
    while (traversedContext != null) {
      SubRoutineStatement sub;
      if ((sub = traversedContext.subroutine()) != null && sub.isSubRoutineEscaping())
        return; 
      if (traversedContext instanceof ExceptionHandlingFlowContext) {
        ExceptionHandlingFlowContext exceptionContext = 
          (ExceptionHandlingFlowContext)traversedContext;
        ReferenceBinding[] caughtExceptions;
        if ((caughtExceptions = exceptionContext.handledExceptions) != Binding.NO_EXCEPTIONS) {
          int caughtCount = caughtExceptions.length;
          boolean[] locallyCaught = new boolean[raisedCount];
          for (int caughtIndex = 0; caughtIndex < caughtCount; caughtIndex++) {
            ReferenceBinding caughtException = caughtExceptions[caughtIndex];
            for (int raisedIndex = 0; raisedIndex < raisedCount; raisedIndex++) {
              TypeBinding raisedException;
              if ((raisedException = raisedExceptions[raisedIndex]) != null) {
                UnconditionalFlowInfo unconditionalFlowInfo;
                FlowInfo exceptionFlow = flowInfo;
                int state = (caughtException == null) ? 
                  -1 : 
                  Scope.compareTypes(raisedException, (TypeBinding)caughtException);
                if (abruptlyExitedLoops != null && caughtException != null && state != 0) {
                  for (int k = 0, abruptlyExitedLoopsCount = abruptlyExitedLoops.size(); k < abruptlyExitedLoopsCount; k++) {
                    LoopingFlowContext loop = abruptlyExitedLoops.get(k);
                    loop.recordCatchContextOfEscapingException(exceptionContext, caughtException, flowInfo);
                  } 
                  unconditionalFlowInfo = FlowInfo.DEAD_END;
                } 
                switch (state) {
                  case -1:
                    exceptionContext.recordHandlingException(
                        caughtException, 
                        unconditionalFlowInfo.unconditionalInits(), 
                        raisedException, 
                        raisedException, 
                        location, 
                        locallyCaught[raisedIndex]);
                    if (!locallyCaught[raisedIndex]) {
                      locallyCaught[raisedIndex] = true;
                      remainingCount--;
                    } 
                    break;
                  case 1:
                    exceptionContext.recordHandlingException(
                        caughtException, 
                        unconditionalFlowInfo.unconditionalInits(), 
                        raisedException, 
                        (TypeBinding)caughtException, 
                        location, 
                        false);
                    break;
                } 
              } 
            } 
          } 
          for (int j = 0; j < raisedCount; j++) {
            if (locallyCaught[j])
              raisedExceptions[j] = null; 
          } 
        } 
        if (exceptionContext.isMethodContext) {
          for (int j = 0; j < raisedCount; j++) {
            TypeBinding raisedException;
            if ((raisedException = raisedExceptions[j]) != null && 
              raisedException.isUncheckedException(false)) {
              remainingCount--;
              raisedExceptions[j] = null;
            } 
          } 
          boolean shouldMergeUnhandledException = exceptionContext instanceof org.eclipse.jdt.internal.compiler.flow.ExceptionInferenceFlowContext;
          if (exceptionContext.associatedNode instanceof AbstractMethodDeclaration) {
            AbstractMethodDeclaration method = (AbstractMethodDeclaration)exceptionContext.associatedNode;
            if (method.isConstructor() && method.binding.declaringClass.isAnonymousType())
              shouldMergeUnhandledException = true; 
          } 
          if (shouldMergeUnhandledException) {
            for (int k = 0; k < raisedCount; k++) {
              TypeBinding raisedException;
              if ((raisedException = raisedExceptions[k]) != null)
                exceptionContext.mergeUnhandledException(raisedException); 
            } 
            return;
          } 
          break;
        } 
      } else if (traversedContext instanceof LoopingFlowContext) {
        if (abruptlyExitedLoops == null)
          abruptlyExitedLoops = new ArrayList<>(5); 
        abruptlyExitedLoops.add(traversedContext);
      } 
      if (remainingCount == 0)
        return; 
      traversedContext.recordReturnFrom(flowInfo.unconditionalInits());
      if (traversedContext instanceof org.eclipse.jdt.internal.compiler.flow.InsideSubRoutineFlowContext) {
        ASTNode node = traversedContext.associatedNode;
        if (node instanceof TryStatement) {
          TryStatement tryStatement = (TryStatement)node;
          flowInfo.addInitializationsFrom((FlowInfo)tryStatement.subRoutineInits);
        } 
      } 
      traversedContext = traversedContext.getLocalParent();
    } 
    for (int i = 0; i < raisedCount; i++) {
      TypeBinding exception;
      if ((exception = raisedExceptions[i]) != null) {
        int j = 0;
        while (true) {
          if (j >= i) {
            scope.problemReporter().unhandledException(exception, location);
            break;
          } 
          if (TypeBinding.equalsEquals(raisedExceptions[j], exception))
            break; 
          j++;
        } 
      } 
    } 
  }
  
  public BranchLabel continueLabel() {
    return null;
  }
  
  public FlowInfo getInitsForFinalBlankInitializationCheck(TypeBinding declaringType, FlowInfo flowInfo) {
    org.eclipse.jdt.internal.compiler.flow.FlowContext current = this;
    FlowInfo inits = flowInfo;
    do {
      if (current instanceof InitializationFlowContext) {
        InitializationFlowContext initializationContext = (InitializationFlowContext)current;
        if (TypeBinding.equalsEquals((TypeBinding)((TypeDeclaration)initializationContext.associatedNode).binding, declaringType))
          return inits; 
        inits = initializationContext.initsBeforeContext;
        current = initializationContext.initializationParent;
      } else if (current instanceof ExceptionHandlingFlowContext) {
        if (current instanceof org.eclipse.jdt.internal.compiler.flow.FieldInitsFakingFlowContext)
          return (FlowInfo)FlowInfo.DEAD_END; 
        ExceptionHandlingFlowContext exceptionContext = (ExceptionHandlingFlowContext)current;
        current = (exceptionContext.initializationParent == null) ? exceptionContext.parent : exceptionContext.initializationParent;
      } else {
        current = current.getLocalParent();
      } 
    } while (current != null);
    throw new IllegalStateException(declaringType.debugName());
  }
  
  public org.eclipse.jdt.internal.compiler.flow.FlowContext getTargetContextForBreakLabel(char[] labelName) {
    org.eclipse.jdt.internal.compiler.flow.FlowContext current = this, lastNonReturningSubRoutine = null;
    while (current != null) {
      if (current.isNonReturningContext())
        lastNonReturningSubRoutine = current; 
      char[] currentLabelName;
      if ((currentLabelName = current.labelName()) != null && 
        CharOperation.equals(currentLabelName, labelName)) {
        ((LabeledStatement)current.associatedNode).bits |= 0x40;
        if (lastNonReturningSubRoutine == null)
          return current; 
        return lastNonReturningSubRoutine;
      } 
      current = current.getLocalParent();
    } 
    return null;
  }
  
  public org.eclipse.jdt.internal.compiler.flow.FlowContext getTargetContextForContinueLabel(char[] labelName) {
    org.eclipse.jdt.internal.compiler.flow.FlowContext current = this;
    org.eclipse.jdt.internal.compiler.flow.FlowContext lastContinuable = null;
    org.eclipse.jdt.internal.compiler.flow.FlowContext lastNonReturningSubRoutine = null;
    while (current != null) {
      if (current.isNonReturningContext()) {
        lastNonReturningSubRoutine = current;
      } else if (current.isContinuable()) {
        lastContinuable = current;
      } 
      char[] currentLabelName;
      if ((currentLabelName = current.labelName()) != null && CharOperation.equals(currentLabelName, labelName)) {
        ((LabeledStatement)current.associatedNode).bits |= 0x40;
        if (lastContinuable != null && 
          current.associatedNode.concreteStatement() == lastContinuable.associatedNode) {
          if (lastNonReturningSubRoutine == null)
            return lastContinuable; 
          return lastNonReturningSubRoutine;
        } 
        return NotContinuableContext;
      } 
      current = current.getLocalParent();
    } 
    return null;
  }
  
  public org.eclipse.jdt.internal.compiler.flow.FlowContext getTargetContextForDefaultBreak() {
    org.eclipse.jdt.internal.compiler.flow.FlowContext current = this, lastNonReturningSubRoutine = null;
    while (current != null) {
      if (current.isNonReturningContext())
        lastNonReturningSubRoutine = current; 
      if (current.isBreakable() && current.labelName() == null) {
        if (lastNonReturningSubRoutine == null)
          return current; 
        return lastNonReturningSubRoutine;
      } 
      current = current.getLocalParent();
    } 
    return null;
  }
  
  public org.eclipse.jdt.internal.compiler.flow.FlowContext getTargetContextForDefaultContinue() {
    org.eclipse.jdt.internal.compiler.flow.FlowContext current = this, lastNonReturningSubRoutine = null;
    while (current != null) {
      if (current.isNonReturningContext())
        lastNonReturningSubRoutine = current; 
      if (current.isContinuable()) {
        if (lastNonReturningSubRoutine == null)
          return current; 
        return lastNonReturningSubRoutine;
      } 
      current = current.getLocalParent();
    } 
    return null;
  }
  
  public org.eclipse.jdt.internal.compiler.flow.FlowContext getInitializationContext() {
    return null;
  }
  
  public org.eclipse.jdt.internal.compiler.flow.FlowContext getLocalParent() {
    if (this.associatedNode instanceof AbstractMethodDeclaration || this.associatedNode instanceof TypeDeclaration || this.associatedNode instanceof org.eclipse.jdt.internal.compiler.ast.LambdaExpression)
      return null; 
    return this.parent;
  }
  
  public String individualToString() {
    return "Flow context";
  }
  
  public FlowInfo initsOnBreak() {
    return (FlowInfo)FlowInfo.DEAD_END;
  }
  
  public UnconditionalFlowInfo initsOnReturn() {
    return FlowInfo.DEAD_END;
  }
  
  public boolean isBreakable() {
    return false;
  }
  
  public boolean isContinuable() {
    return false;
  }
  
  public boolean isNonReturningContext() {
    return false;
  }
  
  public boolean isSubRoutine() {
    return false;
  }
  
  public char[] labelName() {
    return null;
  }
  
  public void markFinallyNullStatus(LocalVariableBinding local, int nullStatus) {
    if (this.initsOnFinally == null)
      return; 
    if (this.conditionalLevel == -1)
      return; 
    if (this.conditionalLevel == 0) {
      this.initsOnFinally.markNullStatus(local, nullStatus);
      return;
    } 
    UnconditionalFlowInfo newInfo = this.initsOnFinally.unconditionalCopy();
    newInfo.markNullStatus(local, nullStatus);
    this.initsOnFinally = (FlowInfo)this.initsOnFinally.mergedWith(newInfo);
  }
  
  public void mergeFinallyNullInfo(FlowInfo flowInfo) {
    if (this.initsOnFinally == null)
      return; 
    if (this.conditionalLevel == -1)
      return; 
    if (this.conditionalLevel == 0) {
      this.initsOnFinally.addNullInfoFrom(flowInfo);
      return;
    } 
    this.initsOnFinally = (FlowInfo)this.initsOnFinally.mergedWith(flowInfo.unconditionalCopy());
  }
  
  public void recordAbruptExit() {
    if (this.conditionalLevel > -1) {
      this.conditionalLevel++;
      if (!(this instanceof ExceptionHandlingFlowContext) && this.parent != null)
        this.parent.recordAbruptExit(); 
    } 
  }
  
  public void recordBreakFrom(FlowInfo flowInfo) {}
  
  public void recordBreakTo(org.eclipse.jdt.internal.compiler.flow.FlowContext targetContext) {}
  
  public void recordContinueFrom(org.eclipse.jdt.internal.compiler.flow.FlowContext innerFlowContext, FlowInfo flowInfo) {}
  
  public boolean recordExitAgainstResource(BlockScope scope, FlowInfo flowInfo, FakedTrackingVariable trackingVar, ASTNode reference) {
    return false;
  }
  
  protected void recordProvidedExpectedTypes(TypeBinding providedType, TypeBinding expectedType, int nullCount) {
    if (nullCount == 0) {
      this.providedExpectedTypes = new TypeBinding[5][];
    } else if (this.providedExpectedTypes == null) {
      int size = 5;
      for (; size <= nullCount; size *= 2);
      this.providedExpectedTypes = new TypeBinding[size][];
    } else if (nullCount >= this.providedExpectedTypes.length) {
      int oldLen = this.providedExpectedTypes.length;
      System.arraycopy(this.providedExpectedTypes, 0, 
          this.providedExpectedTypes = new TypeBinding[nullCount * 2][], 0, oldLen);
    } 
    (new TypeBinding[2])[0] = providedType;
    (new TypeBinding[2])[1] = expectedType;
    this.providedExpectedTypes[nullCount] = new TypeBinding[2];
  }
  
  protected boolean recordFinalAssignment(VariableBinding variable, Reference finalReference) {
    return true;
  }
  
  protected void recordNullReference(LocalVariableBinding local, ASTNode location, int checkType, FlowInfo nullInfo) {}
  
  public void recordUnboxing(Scope scope, Expression expression, int nullStatus, FlowInfo flowInfo) {
    checkUnboxing(scope, expression, flowInfo);
  }
  
  protected void checkUnboxing(Scope scope, Expression expression, FlowInfo flowInfo) {
    int status = expression.nullStatus(flowInfo, this);
    if ((status & 0x2) != 0) {
      scope.problemReporter().nullUnboxing((ASTNode)expression, expression.resolvedType);
      return;
    } 
    if ((status & 0x10) != 0) {
      scope.problemReporter().potentialNullUnboxing((ASTNode)expression, expression.resolvedType);
      return;
    } 
    if ((status & 0x4) != 0)
      return; 
    if (this.parent != null)
      this.parent.recordUnboxing(scope, expression, 1, flowInfo); 
  }
  
  public void recordReturnFrom(UnconditionalFlowInfo flowInfo) {}
  
  public void recordSettingFinal(VariableBinding variable, Reference finalReference, FlowInfo flowInfo) {
    if ((flowInfo.tagBits & 0x1) == 0) {
      org.eclipse.jdt.internal.compiler.flow.FlowContext context = this;
      while (context != null && 
        context.recordFinalAssignment(variable, finalReference))
        context = context.getLocalParent(); 
    } 
  }
  
  public void recordUsingNullReference(Scope scope, LocalVariableBinding local, ASTNode location, int checkType, FlowInfo flowInfo) {
    Expression reference;
    if ((flowInfo.tagBits & 0x3) != 0 || 
      flowInfo.isDefinitelyUnknown(local))
      return; 
    checkType |= this.tagBits & 0x1000;
    int checkTypeWithoutHideNullWarning = checkType & 0xFFFF0FFF;
    switch (checkTypeWithoutHideNullWarning) {
      case 256:
      case 512:
        if (flowInfo.isDefinitelyNonNull(local)) {
          if (checkTypeWithoutHideNullWarning == 512) {
            if ((checkType & 0x1000) == 0)
              scope.problemReporter().localVariableRedundantCheckOnNonNull(local, location); 
            flowInfo.initsWhenFalse().setReachMode(2);
          } else {
            scope.problemReporter().localVariableNonNullComparedToNull(local, location);
            flowInfo.initsWhenTrue().setReachMode(2);
          } 
          return;
        } 
        if (flowInfo.cannotBeDefinitelyNullOrNonNull(local))
          return; 
      case 257:
      case 513:
      case 769:
      case 1025:
        reference = (Expression)location;
        if (flowInfo.isDefinitelyNull(local)) {
          switch (checkTypeWithoutHideNullWarning & 0xFFFF0F00) {
            case 256:
              if ((checkTypeWithoutHideNullWarning & 0xFF) == 1 && (reference.implicitConversion & 0x400) != 0) {
                scope.problemReporter().localVariableNullReference(local, (ASTNode)reference);
                return;
              } 
              if ((checkType & 0x1000) == 0)
                scope.problemReporter().localVariableRedundantCheckOnNull(local, (ASTNode)reference); 
              flowInfo.initsWhenFalse().setReachMode(2);
              return;
            case 512:
              if ((checkTypeWithoutHideNullWarning & 0xFF) == 1 && (reference.implicitConversion & 0x400) != 0) {
                scope.problemReporter().localVariableNullReference(local, (ASTNode)reference);
                return;
              } 
              scope.problemReporter().localVariableNullComparedToNonNull(local, (ASTNode)reference);
              flowInfo.initsWhenTrue().setReachMode(2);
              return;
            case 768:
              scope.problemReporter().localVariableRedundantNullAssignment(local, (ASTNode)reference);
              return;
            case 1024:
              scope.problemReporter().localVariableNullInstanceof(local, (ASTNode)reference);
              return;
          } 
          break;
        } 
        if (flowInfo.isPotentiallyNull(local)) {
          switch (checkTypeWithoutHideNullWarning & 0xFFFF0F00) {
            case 256:
              if ((checkTypeWithoutHideNullWarning & 0xFF) == 1 && (reference.implicitConversion & 0x400) != 0) {
                scope.problemReporter().localVariablePotentialNullReference(local, (ASTNode)reference);
                return;
              } 
              break;
            case 512:
              if ((checkTypeWithoutHideNullWarning & 0xFF) == 1 && (reference.implicitConversion & 0x400) != 0) {
                scope.problemReporter().localVariablePotentialNullReference(local, (ASTNode)reference);
                return;
              } 
              break;
          } 
          break;
        } 
        if (flowInfo.cannotBeDefinitelyNullOrNonNull(local))
          return; 
        break;
      case 3:
        if (flowInfo.isDefinitelyNull(local)) {
          scope.problemReporter().localVariableNullReference(local, location);
          return;
        } 
        if (flowInfo.isPotentiallyNull(local)) {
          if (local.type.isFreeTypeVariable()) {
            scope.problemReporter().localVariableFreeTypeVariableReference(local, location);
            return;
          } 
          scope.problemReporter().localVariablePotentialNullReference(local, location);
          return;
        } 
        break;
    } 
    if (this.parent != null)
      this.parent.recordUsingNullReference(scope, local, location, checkType, 
          flowInfo); 
  }
  
  void removeFinalAssignmentIfAny(Reference reference) {}
  
  public SubRoutineStatement subroutine() {
    return null;
  }
  
  public String toString() {
    StringBuffer buffer = new StringBuffer();
    org.eclipse.jdt.internal.compiler.flow.FlowContext current = this;
    int parentsCount = 0;
    while ((current = current.parent) != null)
      parentsCount++; 
    org.eclipse.jdt.internal.compiler.flow.FlowContext[] parents = new org.eclipse.jdt.internal.compiler.flow.FlowContext[parentsCount + 1];
    current = this;
    int index = parentsCount;
    while (index >= 0) {
      parents[index--] = current;
      current = current.parent;
    } 
    for (int i = 0; i < parentsCount; i++) {
      for (int k = 0; k < i; k++)
        buffer.append('\t'); 
      buffer.append(parents[i].individualToString()).append('\n');
    } 
    buffer.append('*');
    for (int j = 0; j < parentsCount + 1; j++)
      buffer.append('\t'); 
    buffer.append(individualToString()).append('\n');
    return buffer.toString();
  }
  
  public void recordNullityMismatch(BlockScope currentScope, Expression expression, TypeBinding providedType, TypeBinding expectedType, FlowInfo flowInfo, int nullStatus, NullAnnotationMatching annotationStatus) {
    if (providedType == null)
      return; 
    if (expression.localVariableBinding() != null) {
      org.eclipse.jdt.internal.compiler.flow.FlowContext currentContext = this;
      while (currentContext != null) {
        int isInsideAssert = 0;
        if ((this.tagBits & 0x1000) != 0)
          isInsideAssert = 4096; 
        if (currentContext.internalRecordNullityMismatch(expression, providedType, flowInfo, nullStatus, expectedType, 0x80 | isInsideAssert))
          return; 
        currentContext = currentContext.parent;
      } 
    } 
    if (annotationStatus != null) {
      currentScope.problemReporter().nullityMismatchingTypeAnnotation(expression, providedType, expectedType, annotationStatus);
    } else {
      currentScope.problemReporter().nullityMismatch(expression, providedType, expectedType, nullStatus, 
          currentScope.environment().getNonNullAnnotationName());
    } 
  }
  
  protected boolean internalRecordNullityMismatch(Expression expression, TypeBinding providedType, FlowInfo flowInfo, int nullStatus, TypeBinding expectedType, int checkType) {
    return false;
  }
}
